/**
 * @file   juliuslib.h
 * 
 * <EN>
 * @brief  Library top header
 * </EN>
 * 
 * <JA>
 * @brief  ライブラリ使用時のトップヘッダー
 * </JA>
 * 
 * @author Akinobu Lee
 * @date   Tue Nov  6 22:42:12 2007
 * 
 * $Revision: 1.4 $
 * 
 */
/*
 * Copyright (c) 1991-2012 Kawahara Lab., Kyoto University
 * Copyright (c) 2000-2005 Shikano Lab., Nara Institute of Science and Technology
 * Copyright (c) 2005-2012 Julius project team, Nagoya Institute of Technology
 * All rights reserved
 */

#include <julius/julius.h>
